"""Parser module"""

from collections.abc import Iterable
from json import JSONDecodeError, loads
from logging import getLogger
from re import Pattern

from .structure import Device, InterfaceJson, new_device

_LOGGER = getLogger(__name__)


async def parse_raw_lines(
    lines: Iterable[str], regex: Pattern[str]
) -> Iterable[dict[str, str]]:
    """
    Parse the lines using the given regular expression.

    If a line can't be parsed it is logged and skipped in the output.
    We first map all the lines, and after this filter out any None rows.

    Args:
        lines (Iterable[str]): A list of lines to parse
        regex (Pattern[str]): The regex pattern to use on each line
    """

    def _match(line: str) -> dict[str, str] | None:
        """
        Match a single line, will return a None value if no match is made.

        Args:
            line (str): single line to process
        """
        if not line:
            return None

        match = regex.search(line)
        if not match:
            _LOGGER.debug("Could not parse row: %s", line)
            return None

        return match.groupdict()

    return filter(None, map(_match, lines))


def _parse_wl_row(row: dict[str, str], devices: dict[str, Device]) -> None:
    mac = row["mac"].upper()
    devices[mac] = new_device(mac)


def parse_wl(
    data: Iterable[dict[str, str]], devices: dict[str, Device]
) -> None:
    """Parse wl data."""
    _ = list(map(lambda row: _parse_wl_row(row, devices), data))


def _parse_arp_row(row: dict[str, str], devices: dict[str, Device]) -> None:
    mac = row["mac"].upper()
    if mac not in devices:
        devices[mac] = new_device(mac)
    devices[mac].device_data["ip"] = row["ip"]
    devices[mac].interface["id"] = row["interface"]


def _parse_leases(device: dict[str, str], devices: dict[str, Device]) -> None:
    mac = device["mac"].upper()
    host = device.get("host")
    if host is not None and host != "*":
        devices[mac].device_data["name"] = host

    devices[mac].device_data["ip"] = device["ip"]


def _lease_in_devices(
    device: dict[str, str], devices: dict[str, Device]
) -> bool:
    return device["mac"].upper() in devices


def parse_leases(
    data: Iterable[dict[str, str]], devices: dict[str, Device]
) -> None:
    """Parse leases data."""
    _ = list(
        map(
            lambda row: _parse_leases(row, devices),
            filter(lambda row: _lease_in_devices(row, devices), data),
        )
    )


def parse_arp(
    data: Iterable[dict[str, str]], devices: dict[str, Device]
) -> None:
    """Parse arp data."""
    _ = list(map(lambda row: _parse_arp_row(row, devices), data))


def _parse_neigh_row(
    device: dict[str, str], devices: dict[str, Device]
) -> None:
    if not device.get("mac"):
        return
    status = device["status"]
    mac = device["mac"].upper()
    if mac not in devices:
        devices[mac] = new_device(mac)
    devices[mac].device_data["status"] = status
    devices[mac].device_data["ip"] = device.get(
        "ip", devices[mac].device_data["ip"]
    )


def parse_neigh(
    data: Iterable[dict[str, str]], devices: dict[str, Device]
) -> None:
    """Parse neigh data."""
    _ = list(map(lambda row: _parse_neigh_row(row, devices), data))


def _parse_device_clientjson(
    interface_mac: str,
    conn_type: str,
    conn_items: dict[str, dict[str, str | int]],
    dev_mac: str,
    devices: dict[str, Device],
) -> None:
    mac = dev_mac.upper()
    device = conn_items[mac]
    ip = device.get("ip")
    if not isinstance(ip, str):
        ip = None
    rssi = device.get("rssi", None)
    if mac not in devices:
        devices[mac] = new_device(mac)
    devices[mac].device_data["ip"] = ip
    devices[mac].device_data["rssi"] = int(rssi) if rssi else None
    devices[mac].interface["name"] = conn_type
    devices[mac].interface["mac"] = interface_mac


def _map_device_clientjson(
    interface_mac: str,
    conn_type: str,
    conn_items: dict[str, dict[str, str | int]],
    devices: dict[str, Device],
) -> None:
    _ = list(
        map(
            lambda item: _parse_device_clientjson(
                interface_mac,
                conn_type,
                conn_items,
                item,
                devices,
            ),
            conn_items,
        )
    )


def _handle_clientjson(
    interface_mac: str,
    interface: dict[str, dict[str, dict[str, str | int]]],
    devices: dict[str, Device],
) -> None:
    _ = list(
        map(
            lambda conn_type: _map_device_clientjson(
                interface_mac, conn_type, interface[conn_type], devices
            ),
            interface,
        )
    )


def parse_clientjson(data: str, devices: dict[str, Device]) -> None:
    """Parse clientlist.json file"""
    try:
        device_list = InterfaceJson(loads(data))
    except JSONDecodeError:
        _LOGGER.info("clientlist.json is corrupt.")
        return
    _ = list(
        map(
            lambda interface_mac: _handle_clientjson(
                interface_mac, device_list[interface_mac], devices
            ),
            device_list,
        )
    )



"""Unit test asyswrt."""

# pylint: disable=protected-access
# pyright: reportPrivateUsage=false

from unittest.mock import AsyncMock, MagicMock, call, patch

import pytest

from aioasuswrt.asuswrt import AsusWrt
from aioasuswrt.connection import BaseConnection
from aioasuswrt.constant import DEFAULT_DNSMASQ
from aioasuswrt.structure import (
    TEMP_COMMANDS,
    Command,
    Mode,
    Settings,
    TransferRates,
)

from .common import (
    BAD_CLIENTLIST_DATA,
    HOST_DATA,
    LOADAVG_DATA,
    NETDEV_DATA,
    NVRAM_DHCP_DATA,
    NVRAM_DHCP_VALUES,
    NVRAM_FIRMWARE_DATA,
    NVRAM_FIRMWARE_VALUES,
    NVRAM_LABEL_MAC_DATA,
    NVRAM_LABEL_MAC_VALUES,
    NVRAM_MODEL_DATA,
    NVRAM_MODEL_VALUES,
    NVRAM_QOS_DATA,
    NVRAM_QOS_VALUES,
    NVRAM_REBOOT_DATA,
    NVRAM_REBOOT_VALUES,
    NVRAM_WLAN_DATA,
    NVRAM_WLAN_VALUES,
    TEMP_DATA,
    WAKE_DEVICES,
    WAKE_DEVICES_AP,
    WAKE_DEVICES_REACHABLE,
    WAKE_DEVICES_REACHABLE_AND_IP,
    WAKE_DEVICES_REQIRE_IP,
)
from .conftest import successful_get_devices_commands

_BIT_WRAP = 0xFFFFFFFF


@pytest.mark.asyncio
async def test_get_nvram_empty(mocked_wrt: AsusWrt) -> None:
    """Test get_nvram with successful command."""
    mocked_wrt._connection.run_command = AsyncMock(return_value=[])
    assert await mocked_wrt.get_nvram("DHCP") is None


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "command,return_data,result",
    [
        ("DHCP", NVRAM_DHCP_DATA, NVRAM_DHCP_VALUES),
        ("MODEL", NVRAM_MODEL_DATA, NVRAM_MODEL_VALUES),
        ("QOS", NVRAM_QOS_DATA, NVRAM_QOS_VALUES),
        ("REBOOT", NVRAM_REBOOT_DATA, NVRAM_REBOOT_VALUES),
        ("WLAN", NVRAM_WLAN_DATA, NVRAM_WLAN_VALUES),
        ("FIRMWARE", NVRAM_FIRMWARE_DATA, NVRAM_FIRMWARE_VALUES),
        ("LABEL_MAC", NVRAM_LABEL_MAC_DATA, NVRAM_LABEL_MAC_VALUES),
    ],
)
async def test_get_nvram_successful(
    command: str, return_data: str, result: dict[str, str], mocked_wrt: AsusWrt
) -> None:
    """Test get_nvram with successful command."""
    print(command, return_data)
    mocked_wrt._connection.run_command = AsyncMock(return_value=return_data)
    assert result == await mocked_wrt.get_nvram(command)


@pytest.mark.asyncio
async def test_get_current_transfer_first_empty(
    mocked_wrt: AsusWrt,
) -> None:
    """Test get_current_transfer_rates with successful command."""
    mocked_wrt._connection.run_command = AsyncMock(return_value=[])
    assert await mocked_wrt.get_current_transfer_rates() is None


@pytest.mark.asyncio
async def test_get_current_transfer_first_successful(
    mocked_wrt: AsusWrt,
) -> None:
    """Test get_current_transfer_rates with first successful command."""
    mocked_wrt._connection.run_command = AsyncMock(return_value=NETDEV_DATA)
    assert await mocked_wrt.get_current_transfer_rates() == {"rx": 0, "tx": 0}


@pytest.mark.asyncio
async def test_get_current_transfer_second_empty(
    mocked_wrt: AsusWrt,
) -> None:
    """Test get_current_transfer_rates with successful command."""
    mocked_wrt._connection.run_command = AsyncMock(return_value=[])
    with patch("aioasuswrt.asuswrt.time", return_value=120) as mocked_time:
        mocked_wrt._last_transfer_rates_check = 60
        mocked_wrt._transfer_rates = TransferRates(
            -1 + _BIT_WRAP, -1 + _BIT_WRAP
        )
        assert await mocked_wrt.get_current_transfer_rates() is None
        mocked_time.assert_called_once()


@pytest.mark.asyncio
async def test_get_current_transfer_second_successful(
    mocked_wrt: AsusWrt,
) -> None:
    """Test get_current_transfer_rates with second successful command."""
    mocked_wrt._connection.run_command = AsyncMock(return_value=NETDEV_DATA)
    with patch("aioasuswrt.asuswrt.time", return_value=120) as mocked_time:
        mocked_wrt._last_transfer_rates_check = 60
        mocked_wrt._transfer_rates = TransferRates(
            -1 + _BIT_WRAP, -1 + _BIT_WRAP
        )
        assert await mocked_wrt.get_current_transfer_rates() == {
            "rx": 21734767,
            "tx": 13230063,
        }
        mocked_time.assert_called_once()


def _temp_commands_run(command: str) -> list[str] | None:
    if command == "wl -i eth1 phy_tempsense":
        return TEMP_DATA[0]
    if command == "wl -i eth2 phy_tempsense":
        return TEMP_DATA[1]
    if command == "head -c20 /proc/dmu/temperature":
        return TEMP_DATA[2]
    return None


@pytest.mark.asyncio
async def test_get_temperature_first_empty(
    mocked_wrt: AsusWrt,
) -> None:
    """Test get_current_transfer_rates with successful command."""
    mocked_wrt._connection.run_command = AsyncMock(return_value=[])
    assert await mocked_wrt.get_temperature() is None


@pytest.mark.asyncio
async def test_get_temperature_not_numeric(
    mocked_wrt: AsusWrt,
) -> None:
    """Test get_current_transfer_rates with successful command."""
    mocked_wrt._connection.run_command = AsyncMock(
        return_value=["5s (0x3b)\r"]
    )
    assert await mocked_wrt.get_temperature() is None


@pytest.mark.asyncio
async def test_get_temperature_first_successful(
    mocked_wrt: AsusWrt,
) -> None:
    """Test get_current_transfer_rates with successful command."""

    mocked_wrt._connection.run_command = AsyncMock(
        side_effect=_temp_commands_run
    )
    assert await mocked_wrt.get_temperature() == {
        "2.4GHz": 49.5,
        "5.0GHz": 54.5,
        "CPU": 77.0,
    }


@pytest.mark.asyncio
async def test_get_temperature_second_empty(
    mocked_wrt: AsusWrt,
) -> None:
    """Test get_current_transfer_rates with successful command."""
    mocked_wrt._connection.run_command = AsyncMock(return_value=[])
    mocked_wrt._find_temperature_commands = AsyncMock()

    mocked_wrt._temps_commands = {
        "2.4GHz": TEMP_COMMANDS["2.4GHz"][0],
        "5.0GHz": TEMP_COMMANDS["5.0GHz"][0],
        "CPU": TEMP_COMMANDS["CPU"][0],
    }
    assert await mocked_wrt.get_temperature() is None
    mocked_wrt._find_temperature_commands.assert_not_awaited()


@pytest.mark.asyncio
async def test_get_temperature_second_successful(
    mocked_wrt: AsusWrt,
) -> None:
    """Test get_current_transfer_rates with successful command."""
    mocked_wrt._connection.run_command = AsyncMock(
        side_effect=_temp_commands_run
    )
    mocked_wrt._find_temperature_commands = AsyncMock()

    mocked_wrt._temps_commands = {
        "2.4GHz": TEMP_COMMANDS["2.4GHz"][0],
        "5.0GHz": TEMP_COMMANDS["5.0GHz"][0],
        "CPU": TEMP_COMMANDS["CPU"][0],
    }
    assert await mocked_wrt.get_temperature() == {
        "2.4GHz": 49.5,
        "5.0GHz": 54.5,
        "CPU": 77.0,
    }
    mocked_wrt._find_temperature_commands.assert_not_awaited()


@pytest.mark.asyncio
async def test_get_vpn_clients_empty(mocked_wrt: AsusWrt) -> None:
    """Test get_nvram with successful command."""
    mocked_wrt._connection.run_command = AsyncMock(return_value=[])
    assert await mocked_wrt.get_vpn_clients() is None


@pytest.mark.asyncio
async def test_get_connected_devices_router_successful(
    mocked_wrt: AsusWrt,
) -> None:
    """Test for get asuswrt_data in ap mode."""
    mocked_wrt._settings = Settings(mode=Mode.ROUTER)
    mocked_wrt._connection = MagicMock(
        autospec=BaseConnection,
        run_command=AsyncMock(side_effect=successful_get_devices_commands),
    )
    devices = await mocked_wrt.get_connected_devices()
    assert devices == WAKE_DEVICES


@pytest.mark.asyncio
async def test_get_connected_devices_require_ip_router_successful(
    mocked_wrt: AsusWrt,
) -> None:
    """Test for get asuswrt_data in ap mode."""
    mocked_wrt._settings = Settings(mode=Mode.ROUTER, require_ip=True)
    mocked_wrt._connection = MagicMock(
        autospec=BaseConnection,
        run_command=AsyncMock(side_effect=successful_get_devices_commands),
    )
    devices = await mocked_wrt.get_connected_devices()
    assert devices == WAKE_DEVICES_REQIRE_IP


@pytest.mark.asyncio
async def test_get_connected_devices_reachable_router_successful(
    mocked_wrt: AsusWrt,
) -> None:
    """Test for get asuswrt_data in ap mode."""
    mocked_wrt._settings = Settings(mode=Mode.ROUTER)
    mocked_wrt._connection = MagicMock(
        autospec=BaseConnection,
        run_command=AsyncMock(side_effect=successful_get_devices_commands),
    )
    devices = await mocked_wrt.get_connected_devices(True)
    assert devices == WAKE_DEVICES_REACHABLE


@pytest.mark.asyncio
async def test_get_connected_devices_reachable_and_ip_router_successful(
    mocked_wrt: AsusWrt,
) -> None:
    """Test for get asuswrt_data in router mode filter reachable and ip."""
    mocked_wrt._settings = Settings(mode=Mode.ROUTER, require_ip=True)
    mocked_wrt._connection = MagicMock(
        autospec=BaseConnection,
        run_command=AsyncMock(side_effect=successful_get_devices_commands),
    )
    devices = await mocked_wrt.get_connected_devices(True)
    assert devices == WAKE_DEVICES_REACHABLE_AND_IP


@pytest.mark.asyncio
async def test_get_connected_devices_null_values(
    mocked_wrt: AsusWrt,
) -> None:
    """Test for get asuswrt_data with null values."""

    def _no_values(_) -> None:
        """Just a dummy"""

    _cmd_mock = AsyncMock(side_effect=_no_values)
    mocked_wrt._settings = Settings(mode=Mode.ROUTER)
    mocked_wrt._connection = MagicMock(
        autospec=BaseConnection,
        run_command=_cmd_mock,
    )
    devices = await mocked_wrt.get_connected_devices(True)
    assert devices is None
    _cmd_mock.assert_has_calls(
        (
            call(Command.WL),
            call(Command.ARP),
            call(Command.IP_NEIGH),
            call(Command.LEASES.format(DEFAULT_DNSMASQ)),
            call(Command.CLIENTLIST),
        )
    )


@pytest.mark.asyncio
async def test_get_connected_devices_bad_clientlist_data(
    mocked_wrt: AsusWrt,
) -> None:
    """Test for get asuswrt_data with null values."""

    def _bad_clientlist(command: str) -> list[str] | None:
        """Just a dummy"""
        if command == Command.CLIENTLIST:
            return BAD_CLIENTLIST_DATA
        return None

    _cmd_mock = AsyncMock(side_effect=_bad_clientlist)
    mocked_wrt._settings = Settings(mode=Mode.ROUTER)
    mocked_wrt._connection = MagicMock(
        autospec=BaseConnection,
        run_command=_cmd_mock,
    )
    devices = await mocked_wrt.get_connected_devices(True)
    assert devices is None
    _cmd_mock.assert_has_calls(
        (
            call(Command.WL),
            call(Command.ARP),
            call(Command.IP_NEIGH),
            call(Command.LEASES.format(DEFAULT_DNSMASQ)),
            call(Command.CLIENTLIST),
        )
    )


@pytest.mark.asyncio
async def test_get_connected_devices_ap_successful(
    mocked_wrt: AsusWrt,
) -> None:
    """Test for get asuswrt_data in ap mode."""
    _cmd_mock = AsyncMock(side_effect=successful_get_devices_commands)
    mocked_wrt._settings = Settings(mode=Mode.AP)
    mocked_wrt._connection = MagicMock(
        autospec=BaseConnection,
        run_command=_cmd_mock,
    )
    devices = await mocked_wrt.get_connected_devices()
    assert devices == WAKE_DEVICES_AP
    _not_called = False
    try:
        _cmd_mock.assert_any_await(Command.LEASES.format(DEFAULT_DNSMASQ))
    except AssertionError:
        _not_called = True
    assert _not_called


@pytest.mark.asyncio
async def test_get_loadavg_successful(
    mocked_wrt: AsusWrt,
) -> None:
    """Test for get asuswrt_data in ap mode."""
    _cmd_mock = AsyncMock(return_value=LOADAVG_DATA)
    mocked_wrt._connection = MagicMock(
        autospec=BaseConnection,
        run_command=_cmd_mock,
    )
    assert await mocked_wrt.get_loadavg() == {
        "sensor_load_avg1": 0.23,
        "sensor_load_avg15": 0.68,
        "sensor_load_avg5": 0.5,
    }
    _cmd_mock.assert_awaited_once_with(Command.LOADAVG)


@pytest.mark.asyncio
async def test_get_loadavg_null(
    mocked_wrt: AsusWrt,
) -> None:
    """Test for get asuswrt_data in ap mode."""
    _cmd_mock = AsyncMock(return_value=None)
    mocked_wrt._connection = MagicMock(
        autospec=BaseConnection,
        run_command=_cmd_mock,
    )
    assert await mocked_wrt.get_loadavg() is None
    _cmd_mock.assert_awaited_once_with(Command.LOADAVG)


@pytest.mark.asyncio
async def test_get_dns_records_successful(mocked_wrt: AsusWrt) -> None:
    """Test start_vpn_client successfully called."""
    _cmd_mock = AsyncMock(return_value=HOST_DATA)
    mocked_wrt._connection = MagicMock(
        autospec=BaseConnection,
        run_command=_cmd_mock,
    )
    assert await mocked_wrt.get_dns_records() == {
        "127.0.0.1": {
            "host_names": [
                "localhost.localdomain",
                "localhost",
            ],
            "ip": "127.0.0.1",
        },
        "192.168.1.1": {
            "host_names": [
                "RT-AC88U-2780.",
                "RT-AC88U-2780",
                "RT-AC88U-2780.local",
            ],
            "ip": "192.168.1.1",
        },
    }
    _cmd_mock.assert_awaited_once_with(Command.LISTHOSTS)


@pytest.mark.asyncio
async def test_get_dns_records_null(mocked_wrt: AsusWrt) -> None:
    """Test start_vpn_client successfully called."""
    _cmd_mock = AsyncMock(return_value=None)
    mocked_wrt._connection = MagicMock(
        autospec=BaseConnection,
        run_command=_cmd_mock,
    )
    assert await mocked_wrt.get_dns_records() is None
    _cmd_mock.assert_awaited_once_with(Command.LISTHOSTS)


@pytest.mark.asyncio
async def test_total_transfer(mocked_wrt: AsusWrt) -> None:
    """Test start_vpn_client successfully called."""
    _cmd_mock = AsyncMock()
    mocked_wrt._connection = MagicMock(
        autospec=BaseConnection,
        run_command=_cmd_mock,
    )
    mocked_wrt._total_bytes = TransferRates(10, 10)
    assert await mocked_wrt.total_transfer() == {"rx": 10, "tx": 10}
    _cmd_mock.assert_not_awaited()



"""Mappings of structures and typing."""

from enum import StrEnum
from re import Pattern
from re import compile as re_compile
from typing import Callable, NamedTuple, TypeAlias, TypedDict

from .constant import DEFAULT_DNSMASQ, DEFAULT_WAN_INTERFACE

InterfaceJson: TypeAlias = dict[
    str, dict[str, dict[str, dict[str, str | int]]]
]


class AsyncSSHConnectKwargs(TypedDict):
    """Kwargs mapping for the asyncssh.connect method."""

    username: str
    port: int
    server_host_key_algs: list[str]
    password: str | None
    passphrase: str | None
    known_hosts: list[str] | None
    client_keys: list[str] | None


class Interface(TypedDict):
    """
    Interface representation.

    Attributes:
        id (str | None): id of the interface (for example eth0)
        name (str | None): Name of the interface (for example 5g)
        mac (str | None): MAC address for the interface
    """

    id: str | None
    name: str | None
    mac: str | None


class DeviceData(TypedDict):
    """
    Device status representation.

    Attributes:
        ip (str |None): The IP of the device
        name (str | None): The hostname of the device
        status (str | None): The status of the device
        rssi (int | None): Signal strength,
            in a mesh systems this is to closest node.
    """

    ip: str | None
    name: str | None
    status: str | None
    rssi: int | None


class Device(NamedTuple):
    """
    Representation of a connected device.

    Attributes:
        mac (str): The MAC adrress for the device
        device_data (DeviceData): The current status
        interface (Interface): Information about the
            Interface device is connected to
    """

    mac: str
    device_data: DeviceData
    interface: Interface


class TransferRates(NamedTuple):
    """
    Representation of transfer rates.

    Attributes:
        rx (int): Received bytes
        tx (int): Transferred bytes
    """

    rx: int = 0
    tx: int = 0


class ConnectionType(StrEnum):
    """Connection type definition."""

    SSH = "SSH"
    TELNET = "TELNET"


class DNSRecord(TypedDict):
    """DNS record representation."""

    ip: str
    host_names: list[str]


class AuthConfig(TypedDict):
    """
    Authentication configuration

    There are multiple ways to connect to the router,
    we recomend using ssh_key with a passphrase if possible

    Attributes:
        username (str | None): The username to use
        password (str | None): The password to use
            required if no ssh_key is set
        connection_type (ConnectionType | None):
            Defaults to ConnectionType.SSH
        ssh_key (str |None): An optional ssh_key
        passphrase (str |None): An optional passphrase, used for ssh_key
        port (int | None): Defaults to 22 for ssh and 110 for telnet
    """

    username: str | None
    password: str | None
    connection_type: ConnectionType | None
    ssh_key: str | None
    passphrase: str | None
    port: int | None


class Mode(StrEnum):
    """Router modes definition."""

    ROUTER = "router"
    AP = "ap"


class Settings(NamedTuple):
    """
    Settings for communicating with asuswrt router.

    Args:
        require_ip (bool | None): Defaults to False
            if set to True we will not fetch any device we cannot map ip for
        mode (Mode | None): Defaults to Mode.ROUTER
        dnsmasq (str | None): Defaults to "/var/lib/misc"
            Directory where dnsmasq.leases can be found
        wan_interface (str | None): Defaults to eth0
            The name of the WAN interface connection
            used to get external IP and transfer rates
    """

    require_ip: bool = False
    mode: Mode | None = Mode.ROUTER
    dnsmasq: str = DEFAULT_DNSMASQ
    wan_interface: str = DEFAULT_WAN_INTERFACE


class TempCommand(NamedTuple):
    """
    Representation of a temperature command.

    Attributes:
        cli_command (str): The actual cli-command to run
            example "cat /proc/version"
        result_location (int): We .split(" ") the result of the command
            this tells which index of the resulting list to use
        eval_function (Callable): Method we run on the retrieved value
    """

    cli_command: str
    result_location: int
    eval_function: Callable[..., float]


class _Regex(NamedTuple):
    """Regex Mapped to a key."""

    VPN_LIST: Pattern[str] = re_compile(
        (
            r"(?P<description>.+?)>"
            r"(?P<type>.+?)>"
            r"(?P<id>.+?)>"
            r"(?P<username>.*?)>"
            r"(?P<password>.*?)(?:<|$)"
        )
    )
    LEASES: Pattern[str] = re_compile(
        (
            r"\w+\s"
            r"(?P<mac>(([0-9a-fA-F]{2}[:-]){5}([0-9a-fA-F]{2})))\s"
            r"(?P<ip>([0-9]{1,3}[\.]){3}[0-9]{1,3})\s"
            r"(?P<host>([^\s]+))"
        )
    )

    WL: Pattern[str] = re_compile(
        (
            r"\w+\s"
            r"(?P<mac>(([0-9A-F]{2}[:-]){5}([0-9A-F]{2})))"
        )
    )
    HOSTS: Pattern[str] = re_compile(
        r"(?P<ip>(.+[\d][\.][\d]))[\s](?P<hosts>([a-zA-Z].+))"
    )
    NVRAM: str = "^{}=([\\w.\\-/: <>]+)"
    IP_NEIGH: Pattern[str] = re_compile(
        (
            r"(?P<ip>([0-9]{1,3}[\.]){3}[0-9]{1,3}|"
            r"([0-9a-fA-F]{1,4}:){1,7}"
            r"[0-9a-fA-F]{0,4}(:[0-9a-fA-F]{1,4}){1,7})\s"
            r"\w+\s"
            r"\w+.+\s"
            r"(\w+\s(?P<mac>(([0-9a-fA-F]{2}[:-]){5}([0-9a-fA-F]{2}))))?\s"
            r"\s?(router)?"
            r"\s?(nud)?"
            r"(?P<status>(\w+))"
        )
    )
    ARP: Pattern[str] = re_compile(
        (
            r".+\s"
            r"\((?P<ip>([0-9]{1,3}[\.]){3}[0-9]{1,3})\)\s"
            r".+\s"
            r"(?P<mac>(([0-9a-fA-F]{2}[:-]){5}([0-9a-fA-F]{2})))"
            r".+\s"
            r"(?P<interface>([\w.-]+))"
        )
    )


REGEX = _Regex()


class Command(StrEnum):
    """List of commands that we run."""

    GET_PID_OF = "pidof {name}"
    NETDEV = "cat /proc/net/dev"
    RX = "cat /sys/class/net/{}/statistics/rx_bytes"
    TX = "cat /sys/class/net/{}/statistics/tx_bytes"

    MEMINFO = "cat /proc/meminfo"
    LOADAVG = "cat /proc/loadavg"
    LISTHOSTS = "cat /etc/hosts"
    ADDHOST = (
        'cat /etc/hosts | grep -q "{ipaddress} {hostname}" || '
        '(echo "{ipaddress} {hostname}" >> /etc/hosts && '
        "kill -HUP `cat /var/run/dnsmasq.pid`)"
    )
    LEASES = "cat {}/dnsmasq.leases"
    WL = (
        "for dev in `nvram get wl1_vifs && nvram get wl0_vifs && "
        "nvram get wl_ifnames`; do "
        "if type wlanconfig > /dev/null; then "
        "wlanconfig $dev list | awk 'FNR > 1 {print substr($1, 0, 18)}';"
        " else wl -i $dev assoclist; fi; done"
    )
    CLIENTLIST = "cat /tmp/clientlist.json"
    NVRAM = "nvram show |grep '{}'"
    IP_NEIGH = "ip neigh"
    ARP = "arp -n"
    VPN_START = "service start_vpnclient{id}"
    VPN_STOP = "service stop_vpnclient{id}"


class _Nvram(TypedDict):
    """Property mapping for the nvram info command."""

    DHCP: set[str]
    MODEL: set[str]
    QOS: set[str]
    REBOOT: set[str]
    WLAN: set[str]
    GUEST_2G_1: set[str]
    GUEST_2G_2: set[str]
    GUEST_2G_3: set[str]
    WIFI_2G: set[str]
    GUEST_5G_1: set[str]
    GUEST_5G_2: set[str]
    GUEST_5G_3: set[str]
    WIFI_5G: set[str]
    FIRMWARE: set[str]
    LABEL_MAC: set[str]
    VPN: set[str]


_DHCP = {"dhcp_", "_dhcp", "dhcp=", "dhcpd"}
_MODEL = {"model="}
_QOS = {"qos_"}
_REBOOT = {"^reboot_"}
_WLAN = {"wan_"}
_GUEST_2G_1 = {"wl0.1_"}
_GUEST_2G_2 = {"wl0.2_"}
_GUEST_2G_3 = {"wl0.3_"}
_WIFI_2G = {"wl0_"}
_GUEST_5G_1 = {"wl1.1_"}
_GUEST_5G_2 = {"wl1.2_"}
_GUEST_5G_3 = {"wl1.3_"}
_WIFI_5G = {"wl1_"}
_FIRMWARE = {
    "buildinfo",
    "buildno",
    "firmver",
    "firmware_",
    "webs_",
}
_LABEL_MAC = {"label_mac"}

_VPN = {"vpnc_clientlist", "vpn_client"}

Nvram = _Nvram(
    DHCP=_DHCP,
    MODEL=_MODEL,
    QOS=_QOS,
    REBOOT=_REBOOT,
    WLAN=_WLAN,
    GUEST_2G_1=_GUEST_2G_1,
    GUEST_2G_2=_GUEST_2G_2,
    GUEST_2G_3=_GUEST_2G_3,
    WIFI_2G=_WIFI_2G,
    GUEST_5G_1=_GUEST_5G_1,
    GUEST_5G_2=_GUEST_5G_2,
    GUEST_5G_3=_GUEST_5G_3,
    WIFI_5G=_WIFI_5G,
    FIRMWARE=_FIRMWARE,
    LABEL_MAC=_LABEL_MAC,
    VPN=_VPN,
)


def new_device(mac: str) -> Device:
    """
    Initialize a new device.

    Args:
        mac (str): The MAC address of the device
    """
    return Device(
        mac,
        DeviceData(ip=None, name=None, status=None, rssi=None),
        Interface(id=None, name=None, mac=None),
    )


def _eval_divide_two_plus_twenty(val: float) -> float:
    """
    A filter to use on the retrieved float

    val / 2 + 20

    Args:
        val (float): Raw value from router
    """
    return val / 2 + 20


def _eval_divide_one_thousand(val: float) -> float:
    """
    A filter to use on the retrieved float

    val / 1000

    Args:
        val (float): Raw value from router
    """
    if val < 1000:
        raise ValueError(f"Value for dividing with 1000 is to low {val}")
    return val / 1000


def _eval_no_change(val: float) -> float:
    """
    A filter to use on the retrieved float

    Dummy method just returns the val

    Args:
        val (float): Raw value from router
    """
    return val


_TEMP_24_CMDS: list[TempCommand] = [
    TempCommand(
        "wl -i eth1 phy_tempsense",
        0,
        _eval_divide_two_plus_twenty,
    ),
    TempCommand(
        "wl -i eth5 phy_tempsense",
        0,
        _eval_divide_two_plus_twenty,
    ),
]
_TEMP_5_CMDS: list[TempCommand] = [
    TempCommand(
        "wl -i eth2 phy_tempsense",
        0,
        _eval_divide_two_plus_twenty,
    ),
    TempCommand(
        "wl -i eth6 phy_tempsense",
        0,
        _eval_divide_two_plus_twenty,
    ),
]
_TEMP_CPU_CMDS: list[TempCommand] = [
    TempCommand("head -c20 /proc/dmu/temperature", 2, _eval_no_change),
    TempCommand(
        "head -c5 /sys/class/thermal/thermal_zone0/temp",
        0,
        _eval_divide_one_thousand,
    ),
]

TEMP_COMMANDS: dict[str, list[TempCommand]] = {
    "2.4GHz": _TEMP_24_CMDS,
    "5.0GHz": _TEMP_5_CMDS,
    "CPU": _TEMP_CPU_CMDS,
}



"""AioAsusWrt helpers."""

from collections.abc import Iterable
from copy import deepcopy


def empty_iter(iterable: Iterable[str]) -> bool:
    """Checks if an iterator is empty, without consuming."""
    try:
        _ = next(iter(deepcopy(iterable)))
        return False
    except StopIteration:
        return True



"""Constant decleration."""

ALLOWED_KEY_HASHES = [
    "ssh-rsa",
    "rsa-sha2-256",
    "rsa-sha2-512",
    "ecdsa-sha2-nistp256",
    "ecdsa-sha2-nistp384",
    "ecdsa-sha2-nistp521",
    "ssh-ed25519",
    "ssh-ed448",
]

NETDEV_FIELDS = [
    "tx_bytes",
    "tx_packets",
    "tx_errs",
    "tx_drop",
    "tx_fifo",
    "tx_frame",
    "tx_compressed",
    "tx_multicast",
    "rx_bytes",
    "rx_packets",
    "rx_errs",
    "rx_drop",
    "rx_fifo",
    "rx_colls",
    "rx_carrier",
    "rx_compressed",
]

VPN_COUNT = 5
DEFAULT_DNSMASQ = "/var/lib/misc"
DEFAULT_WAN_INTERFACE = "eth0"



"""Module for Asuswrt class."""

from collections.abc import Iterable
from logging import getLogger
from re import finditer, split
from time import time
from typing import cast, final

from .connection import BaseConnection, create_connection
from .constant import NETDEV_FIELDS, VPN_COUNT
from .helpers import empty_iter
from .parsers import (
    parse_arp,
    parse_clientjson,
    parse_leases,
    parse_neigh,
    parse_raw_lines,
    parse_wl,
)
from .structure import (
    REGEX,
    TEMP_COMMANDS,
    AuthConfig,
    Command,
    Device,
    DNSRecord,
    Mode,
    Nvram,
    Settings,
    TempCommand,
    TransferRates,
)

_LOGGER = getLogger(__name__)

_BIT_WRAP = 0xFFFFFFFF


async def _run_temp_command(
    api: BaseConnection, command: TempCommand
) -> float | None:
    command_result: list[str] | None = await api.run_command(
        str(command.cli_command)
    )

    if not command_result:
        return None
    try:
        result = command_result[0].split(" ")[int(command.result_location)]
    except IndexError:
        return None

    if not result or result and not result.isnumeric():
        return None

    return float(command.eval_function(float(result)))


async def _filter_dev_list(
    devices: dict[str, Device], reachable: bool, require_ip: bool
) -> dict[str, Device] | None:
    def _reachable(reachable: bool, dev: Device) -> bool:
        return (
            not reachable
            or reachable
            and dev.device_data.get("status") in ["REACHABLE", "STALE", None]
        )

    def _ip_check(dev: Device, require_ip: bool) -> bool:
        return (
            not require_ip
            or require_ip
            and dev.device_data.get("ip") is not None
        )

    return (
        {
            key: dev
            for key, dev in devices.items()
            if _reachable(reachable, dev) and _ip_check(dev, require_ip)
        }
        if len(devices) > 0
        else None
    )


async def _get_wl(
    connection: BaseConnection, devices: dict[str, Device]
) -> dict[str, Device]:
    """
    Add devices fount in wl.

    Args:
        devices (dict[str, Device]): Currently known device list
    """
    _LOGGER.info("get_wl")

    lines = await connection.run_command(Command.WL)
    if not lines:
        return devices

    parse_wl(await parse_raw_lines(lines, REGEX.WL), devices)
    _LOGGER.info("There are %s devices found in wl", len(devices))
    return devices


async def _get_arp(
    connection: BaseConnection, devices: dict[str, Device]
) -> dict[str, Device]:
    """
    Add devices found in arp.

    Args:
        devices (dict[str, Device]): Currently known device list
    """
    _LOGGER.info("get_arp")
    lines = await connection.run_command(Command.ARP)
    if not lines:
        return devices
    parse_arp(await parse_raw_lines(lines, REGEX.ARP), devices)
    _LOGGER.info("There are %s devices found in arp", len(devices))
    return devices


async def _get_leases(
    connection: BaseConnection, devices: dict[str, Device], dnsmasq: str
) -> dict[str, Device]:
    """
    Add devices found in leases.

    Args:
        devices (dict[str, Device]): Currently known device list
    """
    _LOGGER.info("get_leases")

    lines: Iterable[str] | None = await connection.run_command(
        Command.LEASES.format(dnsmasq)
    )
    if not lines:
        return devices

    lines = filter(lambda line: not line.startswith("duid "), lines)
    if empty_iter(lines):
        return devices

    parse_leases(await parse_raw_lines(lines, REGEX.LEASES), devices)
    _LOGGER.info("There are %s devices found after leases", len(devices))
    return devices


async def _get_neigh(
    connection: BaseConnection, devices: dict[str, Device]
) -> dict[str, Device]:
    """
    Add devices found in neigh.

    Args:
        devices (dict[str, Device]): Currently known device list
    """
    _LOGGER.info("get_neigh")

    lines = await connection.run_command(Command.IP_NEIGH)
    if not lines:
        return devices

    parse_neigh(await parse_raw_lines(lines, REGEX.IP_NEIGH), devices)
    _LOGGER.info("There are %s devices found in neigh", len(devices))
    return devices


async def _get_clientjson(
    connection: BaseConnection, devices: dict[str, Device]
) -> dict[str, Device]:
    """
    Filter devices list using 'clientlist.json' file if available.

    Args:
        devices (dict[str, Device]): Currently known device list
    """
    _LOGGER.info("filter_dev_list")
    lines = await connection.run_command(Command.CLIENTLIST)
    if not lines:
        return devices

    parse_clientjson(lines[0], devices)
    _LOGGER.debug(
        "There are %s devices found after clientlist.json",
        len(devices),
    )
    return devices


@final
class AsusWrt:
    """This is the main interface class."""

    def __init__(
        self,
        host: str,
        auth_config: AuthConfig,
        settings: Settings | None = None,
    ) -> None:
        """
        Initiate the AsusWrt class, and the connection interface.

        Args:
            host (str): IP or hostname for the router
            auth_config (AuthConfig): The authentication configuration
            settings (Settings | None): Optional aioasuswrt settings
        """
        self._settings = settings or Settings()
        self._transfer_rates: TransferRates | None = None
        self._total_bytes = TransferRates()
        self._last_transfer_rates_check = time()
        self._temps_commands: dict[str, TempCommand] | None = None
        self._connection = create_connection(
            host,
            auth_config,
        )

    async def connect(self) -> None:
        """Connect to router."""
        await self._connection.connect()

    @property
    def wan_interface(self) -> str:
        """Wan interface property."""
        return str(self._settings.wan_interface)

    async def get_nvram(
        self, parameter_to_fetch: str
    ) -> dict[str, str] | None:
        """
        Get nvram value.

        Args:
            parameter_to_fetch (str): The parameter we are targeting to fetch.
        """
        target: str = r"\|".join(Nvram.get(parameter_to_fetch, set()))
        cmd = Command.NVRAM.format(target)
        print(cmd)
        data = await self._connection.run_command(cmd)

        if not data:
            _LOGGER.warning("Cant fetch Nvram")
            return None

        values: Iterable[list[str]] = filter(
            lambda row: len(row) > 1 and row[1] not in ["", None],
            map(lambda string: string.split("="), data),
        )

        return dict(list(values))

    async def get_connected_devices(
        self,
        reachable: bool = False,
    ) -> dict[str, Device] | None:
        """
        Retrieve devices and as much info as possible.

        If Settings.require_ip is `True,
            we filter out all devices that do not have ip mapped

        Calls various commands on the router and returns the superset of all
        responses. Some commands will not work on some routers.

        Args:
            reachable (bool): If true,
                filter out all devices that have not REACHABLE as status

        """
        connection = self._connection
        devices: dict[str, Device] = await _get_wl(connection, {})
        _LOGGER.debug(devices)
        devices = await _get_arp(connection, devices)
        _LOGGER.debug(devices)
        devices = await _get_neigh(connection, devices)
        _LOGGER.debug(devices)
        if self._settings.mode != Mode.AP:
            devices = await _get_leases(
                connection, devices, self._settings.dnsmasq
            )
            _LOGGER.debug(devices)
        devices = await _get_clientjson(connection, devices)
        return await _filter_dev_list(
            devices, reachable, self._settings.require_ip
        )

    async def get_current_transfer_rates(
        self,
    ) -> dict[str, int] | None:
        """Get current transfer rates calculated in per second in bytes."""
        _now = time()
        delay = _now - self._last_transfer_rates_check
        self._last_transfer_rates_check = _now

        net_dev_lines = await self._connection.run_command(Command.NETDEV)
        if not net_dev_lines:
            _LOGGER.info("Cannot run netdev command (for transfer rates)")
            return None

        def handle32bitwrap(v: int) -> int:
            return v if v > 0 else v + _BIT_WRAP

        def _add_if_match(line: str) -> TransferRates | None:
            parts = split(r"[\s:]+", line.strip())
            if parts[0] in ["eth0", "vlan1"]:
                return TransferRates(
                    handle32bitwrap(int(parts[1])),
                    handle32bitwrap(int(parts[9])),
                )
            return None

        eth, vlan = list(filter(None, map(_add_if_match, net_dev_lines[2:])))

        inetrx = handle32bitwrap(eth.rx - vlan.rx)
        inettx = handle32bitwrap(eth.tx - vlan.tx)

        rx = tx = 0
        if self._transfer_rates:
            rx = int(handle32bitwrap(inetrx - self._transfer_rates.rx) / delay)
            tx = int(handle32bitwrap(inettx - self._transfer_rates.tx) / delay)
            self._total_bytes = TransferRates(
                handle32bitwrap(
                    self._total_bytes.rx + inetrx - self._transfer_rates.rx
                ),
                handle32bitwrap(
                    self._total_bytes.tx + inettx - self._transfer_rates.tx
                ),
            )

        self._transfer_rates = TransferRates(inetrx, inettx)
        return cast(
            dict[str, int],
            TransferRates(rx, tx)._asdict(),
        )

    async def total_transfer(self) -> dict[str, int]:
        """Total transfer."""
        return cast(dict[str, int], self._total_bytes._asdict())

    async def get_loadavg(self) -> dict[str, float] | None:
        """Get loadavg."""

        _loadavg = await self._connection.run_command(Command.LOADAVG)
        if not _loadavg:
            return None

        loadavg = list(
            map(
                float,
                filter(None, _loadavg[0].split(" ")[0:3]),
            )
        )
        if not len(loadavg) >= 3:
            return None

        return {
            "sensor_load_avg1": loadavg[0],
            "sensor_load_avg5": loadavg[1],
            "sensor_load_avg15": loadavg[2],
        }

    async def get_dns_records(self) -> dict[str, DNSRecord] | None:
        """Get a list of all dns records in hosts file."""
        ret: dict[str, DNSRecord] = {}

        lines = await self._connection.run_command(Command.LISTHOSTS)
        if not lines:
            return None

        def _to_record(row: dict[str, str]) -> None:
            ip = row["ip"]
            hosts = row["hosts"].split(" ")
            if ip in ret:
                ret[ip]["host_names"] += hosts
                return
            ret[ip] = DNSRecord(ip=row["ip"], host_names=hosts)

        data = await parse_raw_lines(lines, REGEX.HOSTS)
        _ = list(map(_to_record, data))
        return ret

    async def add_dns_record(
        self, hostname: str, ipaddress: str
    ) -> list[str] | None:
        """
        Add record to /etc/hosts and HUP dnsmask to catch this record.

        Args:
            hostname (str): Hostname to add
            ipaddress (str): IP address to add
        """
        _records = await self._connection.run_command(
            Command.ADDHOST.format(hostname=hostname, ipaddress=ipaddress)
        )
        if _records:
            return list(_records)
        return None

    async def get_interface_counters(
        self,
    ) -> Iterable[tuple[str, dict[str, int]]] | None:
        """
        Get counters for all network interfaces.

        Returns:
            [('lo', # Interface name
                {'rx_bytes': 537171783, # Received
                   'rx_carrier': 0,
                   'rx_colls': 0,
                   'rx_compressed': 0,
                   'rx_drop': 0,
                   'rx_errs': 0,
                   'rx_fifo': 0,
                   'rx_packets': 2834938,
                   'tx_bytes': 537171783, # Transferred
                   'tx_compressed': 0,
                   'tx_drop': 0,
                   'tx_errs': 0,
                   'tx_fifo': 0,
                   'tx_frame': 0,
                   'tx_multicast': 0,
                   'tx_packets': 2834938
                })
            ]
        """

        net_dev_lines = await self._connection.run_command(Command.NETDEV)
        if not net_dev_lines:
            return None

        lines = map(
            lambda i: list(filter(lambda j: j != "", i.split(" "))),
            net_dev_lines[2:-1],
        )
        interfaces: Iterable[tuple[str, dict[str, int]]] = map(
            lambda i: (
                i[0][0:-1],
                dict(zip(NETDEV_FIELDS, map(int, i[1:]))),
            ),
            lines,
        )
        return interfaces

    async def _find_temperature_commands(self) -> dict[str, float]:
        """Find which temperature commands work with the router, if any."""
        ret: dict[str, float] = {}
        self._temps_commands = {}

        for interface, commands in TEMP_COMMANDS.items():
            temp_command: TempCommand
            ret[interface] = 0.0
            for temp_command in commands:
                result = await _run_temp_command(
                    self._connection, temp_command
                )
                if not result:
                    continue
                self._temps_commands[interface] = temp_command
                ret[interface] = result
                break
        return ret

    async def get_temperature(self) -> dict[str, float] | None:
        """Get temperature values we can find."""
        result: dict[str, float] = {}

        if self._temps_commands is None:
            result = await self._find_temperature_commands()

        if result == {} and self._temps_commands:
            for interface, command in self._temps_commands.items():
                temp = await _run_temp_command(self._connection, command)
                if temp:
                    result[interface] = temp
        if result not in [{"2.4GHz": 0.0, "5.0GHz": 0.0, "CPU": 0.0}, {}]:
            return result
        return None

    async def get_vpn_clients(self) -> list[dict[str, str]] | None:
        """Get current vpn clients."""
        data = await self.get_nvram("VPN")

        if not data:
            return None

        vpn_list = data.get("vpnc_clientlist")
        if not vpn_list:
            return None

        vpns: list[dict[str, str]] = []
        for m in finditer(REGEX.VPN_LIST, vpn_list):
            vpn_id = m.group("id")
            pid = await self._connection.run_command(
                Command.GET_PID_OF.format(name=f"vpnclient{vpn_id}")
            )

            vpn = {k: v for k, v in m.groupdict().items() if v}
            vpn_state_key = f"vpn_client{vpn_id}_state"
            vpn_state = int(data.get(vpn_state_key, 0))

            if vpn_state == 0 or not pid:
                vpn["state"] = "off"
            elif vpn_state == 1:
                vpn["state"] = "starting"
            elif vpn_state == 2:
                vpn["state"] = "on"

            vpns.append(vpn)
        if len(vpns) > 0:
            return vpns
        return None

    async def start_vpn_client(self, vpn_id: int) -> list[str] | None:
        """
        Start a vpn client by id.

        Args:
            vpn_id (int): The id of the VPN service to start
        """
        for no in range(VPN_COUNT):
            _ = await self._connection.run_command(
                Command.VPN_STOP.format(id=no + 1)
            )
        ret: list[str] | None = await self._connection.run_command(
            Command.VPN_START.format(id=vpn_id)
        )
        return ret

    async def stop_vpn_client(self, vpn_id: int) -> list[str] | None:
        """
        Stop a vpn client by id.

        Args:
            vpn_id (int): The id of the VPN service to stop
        """
        ret: list[str] | None = await self._connection.run_command(
            Command.VPN_STOP.format(id=vpn_id)
        )
        return ret

    @property
    def is_connected(self) -> bool:
        """Is connected property."""
        return bool(self._connection.is_connected)

    async def disconnect(self) -> None:
        """Disconnect from router."""
        await self._connection.disconnect()
